#include "herbivore.h"
#include "world.h"
#include "util.h"

#include <QVector>

Herbivore::Herbivore(b2World* box2DWorld, const CreatureConfiguration& configuration, World* parent): Creature(box2DWorld, configuration, parent)
{
    b2BodyDef bodydef;
    bodydef.position.Set(configuration.position.x, configuration.position.y);
    bodydef.type=b2_dynamicBody;
    
    m_body = m_box2DWorld->CreateBody(&bodydef);
    m_body->SetUserData(this);
    
    
    m_createShape(configuration.size);
}


Herbivore::Herbivore( const b2Vec2& pos, float size, b2World *box2DWorld, World *parent ) : Creature( pos, size, box2DWorld, parent )
{
    b2BodyDef bodydef;
    bodydef.position.Set(pos.x, pos.y);
    bodydef.type=b2_dynamicBody;
    
    m_body = m_box2DWorld->CreateBody(&bodydef);
    m_body->SetUserData(this);
    
    m_createShape(size);
}

Creature::Type Herbivore::type()
{
    return Creature::HERBIVORE;
}


void Herbivore::Step()
{
    Creature::Step();
    
    bool busy = false;
    
    // someone's nearby
    if( !m_nearByCreatures.isEmpty() )
    {
	if( m_nearByCreatures.first()->type() == PREDATOR )
	{
	    busy = true;
	    // run for your life
	    // where to run
	    b2Vec2 fromTarget = m_body->GetPosition() - getClosestVisible(PREDATOR)->position();
	    float desiredAngle = atan2( fromTarget.y, fromTarget.x );
	    turnTo( desiredAngle );
	    go(100);
	}
    }
    
    // if energy low, check if plants nearby
    if( energyPercentage() < 0.50 )
    {
	Plant *plant = (Plant *) getClosestVisible(PLANT);
	if( plant )
	{
	    busy = true;
	    // go eat
	    b2Vec2 fromTarget = plant->position() - m_body->GetPosition();
	    float desiredAngle = atan2( fromTarget.y, fromTarget.x );
	    turnTo( desiredAngle );
	    go(50);
	}
    }
    
    // if all is well, look for herbivores to reproduce with
    if( canReproduce() )
    {
	Herbivore *herbivore = (Herbivore *) getClosestVisible(HERBIVORE);
	if( herbivore )
	{
	    busy = true;
	    // go eat
	    b2Vec2 fromTarget = herbivore->position() - m_body->GetPosition();
	    float desiredAngle = atan2( fromTarget.y, fromTarget.x );
	    turnTo( desiredAngle );
	    go(50);
	}
    }
    
    
    // touchy feely
    if( !m_touchingCreatures.isEmpty() )
    {
	foreach( Creature *creature, m_touchingCreatures )
	{
	    switch( creature->type() )
	    {
		case PREDATOR:
		    // nothing really we can do about it
		    break;
		    
		case HERBIVORE:
		    // check for reproduction ability
		    if( canReproduce() && creature->canReproduce() )
		    {
			busy = true;
			reproduce( creature );
		    }
		    break;
		    
		case PLANT:
		    // eat
		    if( energyPercentage() < 2 )
		    {
			busy = true;
			creature->shrink( 0.98 );
			m_energy += creature->takeEnergy(this);
		    }
		    break;
		    
		default:
		    break;
	    }
	}
    }
    
    if( !busy )
    {
	explore();
    }
}


void Herbivore::explore()
{
    Creature::explore();
    
    if( energyPercentage() > 2 ) return;
    
    Plant *plant = (Plant *) getClosestVisible(PLANT);
    if( plant )
    {
	b2Vec2 toTarget = plant->position() - m_body->GetPosition();
	float desiredAngle = atan2( toTarget.y, toTarget.x );
	turnTo( desiredAngle );
	go(50);
    }
}

void Herbivore::m_createShape(double size)
{
    const float DEG2RAD = 3.14159/180;
    QVector<b2Vec2> vertices;
    for (int i=0; i < 360; i += 90)
    {
        float degInRad = i*DEG2RAD;
        vertices.append( b2Vec2(cos(degInRad)*size,sin(degInRad)*size) );
    }
    b2PolygonShape shape;
    shape.Set(vertices.data(), vertices.length());

    b2FixtureDef fixturedef;
    fixturedef.shape=&shape;
    fixturedef.density=1.0;
    m_body->CreateFixture(&fixturedef);
    m_body->SetAngularDamping( 0.5 );
    m_body->SetLinearDamping( 0.5 );
    
    createSensor( 5 * size );
}
